@@ -41,8 +41,7 @@ const Unicorn = (() => {
    */
   function getCsrfToken() {
     const csrfElements = document.getElementsByName("csrfmiddlewaretoken");
-
-    if (csrfElements) {
+    if (csrfElements && csrfElements.length > 0) {
       return csrfElements[0].getAttribute("value");
     }
 
@@ -352,15 +351,142 @@ const Unicorn = (() => {
       this.modelEls = [];
       this.errors = {};
       this.poll = {};
-      this.events = [];
       this.actionQueue = [];
-
-      this.modelEvents = {};
+      this.currentActionQueue = null;
       this.actionEvents = {};
+      this.attachedEventTypes = [];
 
       this.init();
       this.refreshEventListeners();
+      this.initPolling();
+    }
 
+    /**
+     * Initializes the Component.
+     */
+    init() {
+      this.root = $(`[unicorn\\:id="${this.id}"]`);
+
+      if (!this.root) {
+        throw Error("No id found");
+      }
+
+      this.refreshChecksum();
+    }
+
+    /**
+     * Adds an action event listener to the document for each type of event (e.g. click, keyup, etc).
+     * Added at the document level because validation errors would sometimes remove the
+     * events when attached directly to the element.
+     */
+    addActionEventListener(eventType) {
+      document.addEventListener(eventType, (event) => {
+        const targetElement = new Element(event.target);
+
+        if (targetElement && targetElement.isUnicorn && !isEmpty(targetElement.action)) {
+          this.actionEvents[eventType].forEach((element) => {
+            // Use isSameNode (not isEqualNode) because we want to check the nodes reference the same object
+            if (targetElement.el.isSameNode(element.el)) {
+              if (!isEmpty(targetElement.model) && targetElement.model.isLazy) {
+                const action = { type: "syncInput", payload: { name: targetElement.model.name, value: targetElement.getValue() } };
+                this.actionQueue.push(action);
+              }
+
+              if (element.action.key) {
+                if (element.action.key === toKebabCase(event.key)) {
+                  this.callMethod(element.action.name);
+                }
+              } else {
+                this.callMethod(element.action.name);
+              }
+            }
+          });
+        }
+      });
+    }
+
+    /**
+     * Adds a model event listener to the element.
+     * @param {Element} element Element that will get the event attached to.
+     * @param {string} eventType Event type to listen for.
+     */
+    addModelEventListener(element, eventType) {
+      element.el.addEventListener(eventType, () => {
+        const action = { type: "syncInput", payload: { name: element.model.name, value: element.getValue() } };
+        this.actionQueue.push(action);
+
+        this.sendMessage(element.model.debounceTime, (excludeElement, err) => {
+          if (err) {
+            console.error(err);
+          } else if (excludeElement) {
+            this.setModelValues(element);
+          } else {
+            this.setModelValues();
+          }
+        });
+      });
+    }
+
+    /**
+     * Sets event listeners on unicorn elements.
+     */
+    refreshEventListeners() {
+      this.actionEvents = {};
+
+      walk(this.root, (el) => {
+        if (el.isSameNode(this.root)) {
+          // Skip the component root element
+          return;
+        }
+
+        const element = new Element(el);
+
+        if (element.isUnicorn) {
+          if (!isEmpty(element.model)) {
+            if (this.modelEls.filter((m) => m.el.isSameNode(element.el)).length === 0) {
+              this.modelEls.push(element);
+              this.addModelEventListener(element, element.model.eventType);
+            }
+          }
+
+          if (!isEmpty(element.action)) {
+            if (this.actionEvents[element.action.eventType]) {
+              this.actionEvents[element.action.eventType].push(element);
+            } else {
+              this.actionEvents[element.action.eventType] = [element];
+
+              if (this.attachedEventTypes.filter((et) => et === element.action.eventType).length === 0) {
+                this.attachedEventTypes.push(element.action.eventType);
+                this.addActionEventListener(element.action.eventType);
+              }
+            }
+          }
+        }
+      });
+    }
+
+    /**
+     * Calls the method for a particular component.
+     */
+    callMethod(methodName, errCallback) {
+      const action = { type: "callMethod", payload: { name: methodName, params: [] } };
+      this.actionQueue.push(action);
+
+      this.sendMessage(-1, (_, err) => {
+        if (err && typeof errCallback === "function") {
+          errCallback(err);
+        } else if (err) {
+          console.error(err);
+        } else {
+          this.setModelValues();
+        }
+      });
+    }
+
+    /**
+     * Sets up polling if it is defined on the component's root.
+     */
+    initPolling() {
       const rootElement = new Element(this.root);
 
       if (rootElement.isUnicorn && !isEmpty(rootElement.poll)) {
@@ -379,122 +505,6 @@ const Unicorn = (() => {
 
         this.startPolling();
       }
-
-      // Add event listeners to the actual element because they seem to stick around
-      // (attaching to document works for most events, but won't work for lazy/blur)
-      Object.keys(this.modelEvents).forEach((eventType) => {
-        this.modelEvents[eventType].forEach((element) => {
-          element.el.addEventListener(eventType, (event) => {
-            const targetElement = new Element(event.target);
-
-            if (targetElement && targetElement.isUnicorn && !isEmpty(targetElement.model)) {
-              // Use isSameNode (not isEqualNode) because we want to check the nodes reference the same object
-              if (targetElement.el.isSameNode(element.el)) {
-                const action = { type: "syncInput", payload: { name: element.model.name, value: element.getValue() } };
-                this.actionQueue.push(action);
-
-                this.sendMessage(element.model.debounceTime, (err) => {
-                  if (err) {
-                    console.error(err);
-                  } else {
-                    this.setModelValues(element);
-                  }
-                });
-              }
-            }
-          });
-        });
-      });
-
-      // Add event listeners at the document level because validation errors can sometimes remove them
-      Object.keys(this.actionEvents).forEach((eventType) => {
-        document.addEventListener(eventType, (event) => {
-          const targetElement = new Element(event.target);
-
-          if (targetElement && targetElement.isUnicorn && !isEmpty(targetElement.action)) {
-            this.actionEvents[eventType].forEach((element) => {
-              // Use isSameNode (not isEqualNode) because we want to check the nodes reference the same object
-              if (targetElement.el.isSameNode(element.el)) {
-                if (element.action.key) {
-                  if (element.action.key === toKebabCase(event.key)) {
-                    this.callMethod(element.action.name);
-                  }
-                } else {
-                  this.callMethod(element.action.name);
-                }
-              }
-            });
-          }
-        });
-      });
-    }
-
-    /**
-     * Initializes the Component.
-     */
-    init() {
-      this.root = $(`[unicorn\\:id="${this.id}"]`);
-
-      if (!this.root) {
-        throw Error("No id found");
-      }
-
-      this.refreshChecksum();
-    }
-
-    /**
-     * Sets event listeners on unicorn elements.
-     */
-    refreshEventListeners() {
-      this.modelEvents = {};
-      this.actionEvents = {};
-
-      walk(this.root, (el) => {
-        if (el.isSameNode(this.root)) {
-          // Skip the component root element
-          return;
-        }
-
-        const element = new Element(el);
-
-        if (element.isUnicorn) {
-          if (!isEmpty(element.model)) {
-            this.modelEls.push(element);
-
-            if (this.modelEvents[element.model.eventType]) {
-              this.modelEvents[element.model.eventType].push(element);
-            } else {
-              this.modelEvents[element.model.eventType] = [element];
-            }
-          }
-
-          if (!isEmpty(element.action)) {
-            if (this.actionEvents[element.action.eventType]) {
-              this.actionEvents[element.action.eventType].push(element);
-            } else {
-              this.actionEvents[element.action.eventType] = [element];
-            }
-          }
-        }
-      });
-    }
-
-    /**
-     * Calls the method for a particular component.
-     */
-    callMethod(methodName, errCallback) {
-      const action = { type: "callMethod", payload: { name: methodName, params: [] } };
-      this.actionQueue.push(action);
-
-      this.sendMessage(-1, (err) => {
-        if (err && typeof errCallback === "function") {
-          errCallback(err);
-        } else if (err) {
-          console.error(err);
-        } else {
-          this.setModelValues();
-        }
-      });
     }
 
     /**
@@ -582,11 +592,25 @@ const Unicorn = (() => {
      */
     sendMessage(debounceTime, callback) {
       function _sendMessage(_component) {
+        // Prevent network call when there isn't an action
+        if (_component.actionQueue.length === 0) {
+          return;
+        }
+
+        // Prevent newtwork call when the action queue gets repeated
+        if (_component.currentActionQueue === _component.actionQueue) {
+          return;
+        }
+
+        // Set the current action queue and clear the action queue in case another event happens
+        _component.currentActionQueue = _component.actionQueue;
+        _component.actionQueue = [];
+
         const body = {
           id: _component.id,
           data: _component.data,
           checksum: _component.checksum,
-          actionQueue: _component.actionQueue,
+          actionQueue: _component.currentActionQueue,
         };
 
         const headers = {
@@ -617,9 +641,6 @@ const Unicorn = (() => {
               throw Error(responseJson.error);
             }
 
-            // Refresh the action queue when a response works
-            _component.actionQueue = [];
-
             // Remove any unicorn validation messages before trying to merge with morphdom
             _component.modelEls.forEach((element) => {
               // Re-initialize element to make sure it is up to date
@@ -674,19 +695,37 @@ const Unicorn = (() => {
               });
             });
 
+            // Check if the current actionQueue contains a callMethod. Prevents excluding
+            // an element from getting a value set because calling a component function can
+            // potentially have side effects which have to be reflected on the current element.
+            let hasCallMethod = false;
+
+            _component.currentActionQueue.forEach((action) => {
+              if (action.type === "callMethod") {
+                hasCallMethod = true;
+              }
+            });
+
+            // Clear the current action queue
+            _component.currentActionQueue = null;
+
             if (callback && typeof callback === "function") {
-              callback();
+              callback(!hasCallMethod, null);
             }
           })
           .catch((err) => {
+            // Make sure to clear the current queues in case of an error
+            _component.actionQueue = [];
+            _component.currentActionQueue = null;
+
             if (callback && typeof callback === "function") {
-              callback(err);
+              callback(null, err);
             }
           });
       }
 
       if (debounceTime === -1) {
-        debounce(_sendMessage, 250, true)(this);
+        debounce(_sendMessage, 250, false)(this);
         // queue(_sendMessage, 250)(this);
       } else {
         debounce(_sendMessage, debounceTime, false)(this);
