@@ -1,6 +1,7 @@
 import hmac
 from functools import wraps
 from typing import Any, Dict, List, Tuple, Union
+from django.forms.forms import Form
 
 import orjson
 import shortuuid
@@ -30,11 +31,12 @@ def handle_error(view_func):
 
 
 def _set_property_from_data(
-    component_or_field: Union[UnicornView, UnicornField, Model], name: str, value
+    component_or_field: Union[UnicornView, UnicornField, Model], name: str, value,
 ) -> None:
     """
     Sets properties on the component based on passed-in data.
     """
+
     if hasattr(component_or_field, name):
         field = getattr(component_or_field, name)
 
@@ -44,7 +46,11 @@ def _set_property_from_data(
                 key_value = value[key]
                 _set_property_from_data(field, key, key_value)
         else:
-            setattr(component_or_field, name, value)
+            if hasattr(component_or_field, "_set_property"):
+                # Can assume that `component_or_field` is a component
+                component_or_field._set_property(name, value)
+            else:
+                setattr(component_or_field, name, value)
 
 
 def _set_property_from_payload(
@@ -62,6 +68,7 @@ def _set_property_from_payload(
 
     property_name = payload.get("name")
     property_value = payload.get("value")
+    component.updating(property_name, property_value)
 
     if property_name is not None and property_value is not None:
         """
@@ -84,7 +91,25 @@ def _set_property_from_payload(
         for (idx, property_name_part) in enumerate(property_name_parts):
             if hasattr(component_or_field, property_name_part):
                 if idx == len(property_name_parts) - 1:
-                    setattr(component_or_field, property_name_part, property_value)
+                    if hasattr(component_or_field, "_set_property"):
+                        # Can assume that `component_or_field` is a component
+                        component_or_field._set_property(
+                            property_name_part, property_value
+                        )
+                    else:
+                        # Handle calling the updating/updated method for nested properties
+                        property_name_snake_case = property_name.replace(".", "_")
+                        updating_function_name = f"updating_{property_name_snake_case}"
+                        updated_function_name = f"updated_{property_name_snake_case}"
+
+                        if hasattr(component, updating_function_name):
+                            getattr(component, updating_function_name)(property_value)
+
+                        setattr(component_or_field, property_name_part, property_value)
+
+                        if hasattr(component, updated_function_name):
+                            getattr(component, updated_function_name)(property_value)
+
                     data_or_dict[property_name_part] = property_value
                 else:
                     component_or_field = getattr(component_or_field, property_name_part)
@@ -97,6 +122,8 @@ def _set_property_from_payload(
                     component_or_field = component_or_field[property_name_part]
                     data_or_dict = data_or_dict.get(property_name_part, {})
 
+    component.updated(property_name, property_value)
+
 
 def _parse_call_method_name(call_method_name: str) -> Tuple[str, List[Any]]:
     """
@@ -241,10 +268,17 @@ def message(request: HttpRequest, component_name: str = None) -> JsonResponse:
     component = UnicornView.create(
         component_id=component_request.id, component_name=component_name
     )
+    validate_all_fields = False
+
+    # Get a copy of the data passed in to determine what fields are updated later
+    original_data = component_request.data.copy()
 
     # Set component properties based on request data
     for (name, value) in component_request.data.items():
         _set_property_from_data(component, name, value)
+    component.hydrate()
+
+    is_reset_called = False
 
     for action in component_request.action_queue:
         action_type = action.get("type")
@@ -256,36 +290,67 @@ def message(request: HttpRequest, component_name: str = None) -> JsonResponse:
             call_method_name = payload.get("name", "")
             assert call_method_name, "Missing 'name' key for callMethod"
 
-            # Handle the special case of the reset action
             if call_method_name == "reset" or call_method_name == "reset()":
+                # Handle the reset special action
                 component = UnicornView.create(
                     component_id=component_request.id,
                     component_name=component_name,
                     use_cache=False,
                 )
+
+                #  Explicitly remove all errors and prevent validation from firing before render()
+                component.errors = {}
+                is_reset_called = True
+            elif call_method_name == "refresh" or call_method_name == "refresh()":
+                # Handle the refresh special action
+                component = UnicornView.create(
+                    component_id=component_request.id,
+                    component_name=component_name,
+                    use_cache=True,
+                )
+            elif call_method_name == "validate" or call_method_name == "validate()":
+                # Handle the validate special action
+                validate_all_fields = True
             elif "=" in call_method_name:
                 call_method_name_split = call_method_name.split("=")
                 property_name = call_method_name_split[0]
                 property_value = _handle_arg(call_method_name_split[1])
 
                 if hasattr(component, property_name):
+                    component.calling(f"set_{property_name}", property_value)
                     setattr(component, property_name, property_value)
+                    component.called(f"set_{property_name}", property_value)
                     component_request.data[property_name] = property_value
             else:
                 (method_name, params) = _parse_call_method_name(call_method_name)
+                component.calling(method_name, params)
                 _call_method_name(component, method_name, params)
+                component.called(method_name, params)
         else:
             raise UnicornViewError(f"Unknown action_type '{action_type}'")
 
-    rendered_component = component.render()
-
     # Re-load frontend context variables to deal with non-serializable properties
     component_request.data = orjson.loads(component.get_frontend_context_variables())
 
+    if not is_reset_called:
+        if validate_all_fields:
+            component.validate()
+        else:
+            model_names_to_validate = []
+
+            for key, value in original_data.items():
+                if value != component_request.data[key]:
+                    model_names_to_validate.append(key)
+
+            component.validate(model_names=model_names_to_validate)
+
+    rendered_component = component.render()
+
     res = {
         "id": component_request.id,
         "dom": rendered_component,
         "data": component_request.data,
+        "errors": component.errors,
     }
 
     return JsonResponse(res)
